from linkify_it import LinkifyIt
import httpx
from bs4 import BeautifulSoup

import re
import io
import PIL.Image

from asyncio import timeout
import traceback

from yobot.interface import ILine, IServer, IPlugin


TOO_LONG = 1024 * 1024 * 16  # 16mb


def human_readable_size(size: int, decimal_places: int = 2) -> str:
    size_ = float(size)
    unit = "B"
    for unit in ["B", "KiB", "MiB", "GiB", "TiB", "PiB"]:
        if size_ < 1024.0 or unit == "PiB":
            break
        size_ /= 1024.0
    return f"{size_:.{decimal_places}f} {unit}"


def handle_html(_response: httpx.Response, content: bytes) -> str:
    soup = BeautifulSoup(content, "html.parser")
    if soup.title is None or soup.title.string is None:
        raise RuntimeError("unable to get title")
    title = soup.title.string
    return re.sub(r"\s+", " ", title).strip()


def handle_image(_response: httpx.Response, content: bytes) -> str:
    buffer = io.BytesIO(content)
    try:
        with PIL.Image.open(buffer) as img:
            return f"Image ({img.format}) ({img.width}x{img.height})"
    except:
        return "Image (unknown)"


def handle_pdf(_response: httpx.Response, _content: bytes) -> str:
    return "pdf"


def handle_unknown(response: httpx.Response) -> str:
    res = ""
    if "content-type" in response.headers:
        res += response.headers["content-type"]
    if res:
        res += " "
    if "content-length" in response.headers:
        res += f'[{human_readable_size(int(response.headers["content-length"]))}]'
    return res


async def aread_max(req: httpx.Response, maxlength: int) -> bytes:
    parts = []
    length = 0
    async for part in req.aiter_bytes():
        parts.append(part)
        length += len(parts)
        if length > maxlength:
            raise RuntimeError("maximum length exceeded")
    req._content = b"".join(parts)
    return req._content


async def generic(url: str) -> str:
    async with httpx.AsyncClient(
        headers={
            "User-Agent": "Yobot/1.0 (mailto:bot@yori.cc)",
            "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/webp,*/*;q=0.8",
            "Accept-Encoding": "gzip, deflate, br",
            "Accept-Language": "en-US,en;q=0.7,nl;q=0.3",
        },
        follow_redirects=True,
    ) as client:
        async with client.stream("GET", url) as r:
            ct = r.headers["content-type"]
            if (
                "content-length" not in r.headers
                or int(r.headers["content-length"]) < TOO_LONG
            ):
                content = await aread_max(r, TOO_LONG)
                if ct.split(";")[0].strip() == "text/html":
                    return handle_html(r, content)
                if ct.startswith("image/"):
                    return handle_image(r, content)
                if ct == "application/pdf":
                    return handle_pdf(r, content)
            return handle_unknown(r)


def get_links(linkify: LinkifyIt, msg: str) -> list[str]:
    matches = linkify.match(msg) or []
    return [m.url for m in matches if m.schema in ("https:", "http:")]


class Plugin(IPlugin):
    def __init__(self, config: dict) -> None:
        super().__init__(config)
        self.linkify = LinkifyIt()

    async def handle(self, bot: IServer, line: ILine) -> None:
        msg = line.params[1]
        async with timeout(10):
            for link in get_links(self.linkify, msg):
                try:
                    res = await self.metadata(link)
                    if len(res) > 400:
                        res = res[:400] + "..."
                except Exception as e:
                    traceback.print_exc()
                    print("error", e)
                else:
                    await bot.reply(line, res)

    async def metadata(self, url: str) -> str:
        if re.match(r"https?://twitter\.com/", url):
            return await self.twitter(url)
        return await self.generic(url)

    async def twitter(self, url: str) -> str:
        url = re.sub(
            r"^https?://twitter\.com/",
            self.config["nitter_domain"] + "/",
            url,
        )
        res = await self.generic(url)
        return re.sub(r"\| nitter$", "| twitter", res)

    async def generic(self, url: str) -> str:
        return await generic(url)
